pub const CFGNAME: &str = "config.ini";
use configparser::ini::Ini;
use eyre::Context;
use std::{fs::File, io::ErrorKind};
use tracing::warn;

fn default_cfg() -> Ini {
    let mut config = Ini::new();
    config.set("core", "worker-threads", Some("4".into()));

    config.set("logging", "log_level", Some("info".into()));
    config.set("logging", "custom_filter", Some("".into()));

    config.set("influx", "socketaddr", Some("127.0.0.1:8086".into()));
    config.set("influx", "https", Some("false".into()));
    config.set("influx", "username", Some("default".into()));
    config.set("influx", "password", Some("".into()));

    config.set("querying", "interval", Some("10".into()));
    config.set("querying", "max-concurrent", Some("5".into()));
    config
}

/// Loads the program configuration from `config.ini` beside the executable.
///
/// Returns an error if the configuration file:
/// * Cannot be parsed
/// * Does not exist, and cannot be created.
/// * Cannot be accessed
pub fn get_settings() -> Result<Ini, eyre::Report> {
    // This sets the configuration file name relative to the exe location.
    let mut path = std::env::current_exe()?;
    path.pop();
    path.push(CFGNAME);

    let mut config = default_cfg();

    // Check if the cfg file exists, and can be written to.
    match File::open(&path) {
        Ok(_) => match config.load(&path) {
            Ok(_) => Ok(config),
            Err(e) => Err(eyre::Report::msg(e)).wrap_err("When loading configuration file"),
        },
        Err(e) => match e.kind() {
            ErrorKind::NotFound => {
                warn!("No configuration file found...");
                config.write(&path)?;
                warn!("Wrote default to '{}'", &path.to_string_lossy());
                Ok(config)
            }
            _ => Err(e)
                .wrap_err_with(|| format!("Unable to access cfg file {}", &path.to_string_lossy())),
        },
    }
}
