//! Methods for interacting with a Mumble/Murmur server.
//!
//! This module doesn't keep any connection state, and calls are blocking.

use crate::{serverlist::QueryServer, ServerResponse};
use byteorder::{BigEndian, ReadBytesExt, WriteBytesExt};
use std::{io::Cursor, net::UdpSocket, time::Duration};

/// Structured data representing Mumble's _Pong_ packet.
///
/// Reference used: https://wiki.mumble.info/wiki/Protocol
#[derive(Debug)]
pub struct MumblePong {
    pub version: String,
    pub ident: u64,
    pub clients: u32,
    pub max_clients: u32,
    pub max_bandwidth: u32,
}

// A tad messy.
impl MumblePong {
    /// Takes a byte buffer and returns a constructed MumblePong.
    fn new(buf: [u8; 24]) -> Self {
        // Since we're reading from an array we know how many bytes the
        // buffer is.  So assuming we keep track of where we are when reading,
        // read_<len> will never Err, and we can unwrap to default losslessly.
        let mut cursor = Cursor::new(buf);
        // First 4 bytes
        // "Can be interpreted as one single int or four signed chars."
        // so we're doing the latter.
        let major_tens = cursor.read_u8().unwrap_or_default() as char; // 1
        let major = cursor.read_u8().unwrap_or_default() as char; // 2
        let minor = cursor.read_u8().unwrap_or_default() as char; // 3
        let patch = cursor.read_u8().unwrap_or_default() as char; // 4
        let version = format!("{}{}.{}.{}", major_tens, major, minor, patch);
        MumblePong {
            version,
            ident: cursor.read_u64::<BigEndian>().unwrap_or_default(), // 12
            clients: cursor.read_u32::<BigEndian>().unwrap_or_default(), // 16
            max_clients: cursor.read_u32::<BigEndian>().unwrap_or_default(), // 20
            max_bandwidth: cursor.read_u32::<BigEndian>().unwrap_or_default(), // 24
        }
    }
}

/// Opens a UDP socket to a mumble server to get its publicly avaliable info.
///
/// # Errors
/// This function mostly returns errors related to sockets and reading therein.
///
/// If the [QueryServer] doesn't respond within 300ms,
/// or rejects the connection.
pub fn query_server<'a>(
    server: &'a QueryServer,
    time_now: &chrono::DateTime<chrono::Utc>,
) -> Result<ServerResponse<'a>, eyre::Error> {
    // If we bind to an ipv6 address and then try to connect to an ipv4 remote,
    // windows can't handle it, even if we stuff the ipv4 address into v6 space
    // so we have to bind locally the corresponding remote stack.
    let socket = match server.socket.is_ipv4() {
        true => UdpSocket::bind("0.0.0.0:0")?,
        false => UdpSocket::bind("[::]:0")?,
    };

    socket.set_read_timeout(Some(Duration::from_millis(300)))?;
    socket.connect(server.socket)?;

    // Mumble is BIG ENDIAN over the wire.

    let mut request = Vec::with_capacity(12);
    // Request Type, 0x0 = Ping
    request.write_u32::<BigEndian>(0x0)?;
    // Magic Number
    request.write_u64::<BigEndian>(0x_20_67_76_69_64_2e_6d_65)?;

    let mut response = [0u8; 24];

    socket.send(&request)?;
    socket.recv(&mut response)?;
    let pong = MumblePong::new(response);
    Ok(ServerResponse {
        server,
        clients: pong.clients,
        time: *time_now,
        map: None,
    })
}
